#include <algorithm>
#include <pybind11/complex.h>
#include <pybind11/numpy.h>
#include <pybind11/stl.h>
#include <set>
#include <tuple>
#include <vector>

#include <ctime>
#include <iostream>

namespace py = pybind11;

class Timer {
public:
  Timer() { clock_gettime(CLOCK_REALTIME, &beg_); }

  double elapsed() {
    clock_gettime(CLOCK_REALTIME, &end_);
    return end_.tv_sec - beg_.tv_sec +
           (end_.tv_nsec - beg_.tv_nsec) / 1000000000.;
  }

  void reset() { clock_gettime(CLOCK_REALTIME, &beg_); }

private:
  timespec beg_, end_;
};

template <typename T>
std::vector<T> vectorSortIntArr(std::vector<std::array<T, 2>> v) {
  std::sort(v.begin(), v.end());
  // double t = tmr.elapsed();
  // tmr.reset();
  auto iter = std::unique(v.begin(), v.end());
  // t = tmr.elapsed();
  // std::cout << t << std::endl;

  size_t len = iter - v.begin();
  std::vector<T> outvec;
  outvec.reserve(len * 2);
  for (auto i = v.begin(); i != iter; ++i) {
    outvec.push_back(i->at(0));
    outvec.push_back(i->at(1));
  }
  return outvec;
}

py::array unique_edges(
    py::array_t<int, py::array::c_style | py::array::forcecast> edges) {

  std::vector<int> cedges(edges.size());
  std::memcpy(cedges.data(), edges.data(), edges.size() * sizeof(int));

  std::vector<std::array<int, 2>> tl;

  tl.reserve(cedges.size());
  for (size_t i = 0; i < cedges.size(); i += 2) {
    tl.push_back({std::min(cedges[i], cedges[i + 1]),
                  std::max(cedges[i], cedges[i + 1])});
  }

  auto u_edges = vectorSortIntArr<int>(std::move(tl));

  int num_edges = u_edges.size();
  ssize_t sint = sizeof(int);
  std::vector<ssize_t> shape = {num_edges / 2, 2};
  std::vector<ssize_t> strides = {sint * 2, sint};
  return py::array(
      py::buffer_info(u_edges.data(), /* data as contiguous array  */
                      sizeof(int),    /* size of one scalar        */
                      py::format_descriptor<int>::format(), /* data type */
                      2,      /* number of dimensions      */
                      shape,  /* shape of the matrix       */
                      strides /* strides for each axis     */
                      ));
}

double l2_norm(const std::array<double, 3> &u) {
  double accum = 0.;
  for (int i = 0; i < 3; ++i) {
    accum += u[i] * u[i];
  }
  return sqrt(accum);
}

std::array<double, 3> cross_product(const std::array<double, 3> &vect_A,
                                    const std::array<double, 3> &vect_B) {
  std::array<double, 3> cross_P;
  cross_P[0] = vect_A[1] * vect_B[2] - vect_A[2] * vect_B[1];
  cross_P[1] = vect_A[2] * vect_B[0] - vect_A[0] * vect_B[2];
  cross_P[2] = vect_A[0] * vect_B[1] - vect_A[1] * vect_B[0];
  return cross_P;
}

double dot_product(const std::array<double, 3> &vect_A,
                   const std::array<double, 3> &vect_B) {
  double product = 0.0;
  for (unsigned int i = 0; i < 3; i++)
    product = product + vect_A[i] * vect_B[i];
  return product;
}

std::vector<double> c_dblock(std::vector<double> &points, double x1, double x2,
                             double y1, double y2, double z1, double z2) {
  // Signed distance function for block with corners (x1,y1,z1), (x2,y2,z2)
  int sz = points.size() / 3;
  std::vector<double> dist;
  dist.resize(sz);
  // return -min(
  //    min(
  //        min(min(min(-z1 + p[:, 2], z2 - p[:, 2]), -y1 + p[:, 1]), y2 - p[:,
  //        1]), -x1 + p[:, 0],
  //    ),
  //    x2 - p[:, 0],
  //)
  for (std::size_t iv = 0; iv < sz; ++iv) {

    double sum1 = -z1 + points[3 * iv + 2];
    double sum2 = z2 - points[3 * iv + 2];
    double sum3 = -y1 + points[3 * iv + 1];
    double sum4 = y2 - points[3 * iv + 1];
    double sum5 = -x1 + points[3 * iv];
    double sum6 = x2 - points[3 * iv];

    double tmp1 = std::min(sum1, sum2);
    double tmp2 = std::min(tmp1, sum3);
    double tmp3 = std::min(tmp2, sum4);
    double tmp4 = std::min(tmp3, sum5);

    dist[iv] = -std::min(tmp4, sum6);
  }
  return dist;
}

py::array dblock_fast(
    py::array_t<double, py::array::c_style | py::array::forcecast> points,
    double x1, double x2, double y1, double y2, double z1, double z2) {

  // check input dimensions
  int num_points = points.size() / 3;
  // allocate std::vector (to pass to the C++ function)
  std::vector<double> cppPts(num_points * 3);

  // copy py::array -> std::vector
  std::memcpy(cppPts.data(), points.data(), 3 * num_points * sizeof(double));

  std::vector<double> dist = c_dblock(cppPts, x1, x2, y1, y2, z1, z2);

  ssize_t sodble = sizeof(double);
  std::vector<ssize_t> shape = {num_points};
  std::vector<ssize_t> strides = {sodble};

  // return 1-D NumPy array
  return py::array(
      py::buffer_info(dist.data(),    /* data as contiguous array  */
                      sizeof(double), /* size of one scalar        */
                      py::format_descriptor<double>::format(), /* data type */
                      1,      /* number of dimensions      */
                      shape,  /* shape of the matrix       */
                      strides /* strides for each axis     */
                      ));
}

std::vector<double> c_drectangle(std::vector<double> &points, double x1,
                                 double x2, double y1, double y2) {
  // Signed distance function for rectangle with corners (x1,y1), (x2,y1)
  int sz = points.size() / 2;
  std::vector<double> dist;
  dist.resize(sz);
  // return -min(min(min(-y1 + p[:, 1], y2 - p[:, 1]), -x1 + p[:, 0]), x2 - p[:,
  // 0])
  for (std::size_t iv = 0; iv < sz; ++iv) {

    double sum1 = -y1 + points[2 * iv + 1];
    double sum2 = y2 - points[2 * iv + 1];
    double sum3 = -x1 + points[2 * iv];
    double sum4 = x2 - points[2 * iv];

    double tmp1 = std::min(sum1, sum2);
    double tmp2 = std::min(tmp1, sum3);
    dist[iv] = -std::min(tmp2, sum4);
  }
  return dist;
}

py::array drectangle_fast(
    py::array_t<double, py::array::c_style | py::array::forcecast> points,
    double x1, double x2, double y1, double y2) {

  // check input dimensions
  int num_points = points.size() / 2;
  // allocate std::vector (to pass to the C++ function)
  std::vector<double> cppPts(num_points * 2);

  // copy py::array -> std::vector
  std::memcpy(cppPts.data(), points.data(), 2 * num_points * sizeof(double));

  std::vector<double> dist = c_drectangle(cppPts, x1, x2, y1, y2);

  ssize_t sodble = sizeof(double);
  std::vector<ssize_t> shape = {num_points};
  std::vector<ssize_t> strides = {sodble};

  // return 1-D NumPy array
  return py::array(
      py::buffer_info(dist.data(),    /* data as contiguous array  */
                      sizeof(double), /* size of one scalar        */
                      py::format_descriptor<double>::format(), /* data type */
                      1,      /* number of dimensions      */
                      shape,  /* shape of the matrix       */
                      strides /* strides for each axis     */
                      ));
}

std::vector<double> c_remove_external_entities2(std::vector<double> &points, std::vector<int> &cells, double x1, double x2, double y1, double y2){

    int num_points = points.size()/2;
    int num_cells = cells.size()/3;
    std::vector<double> tmp;
    tmp.resize(num_cells*3*2);

    int kount = 0;
    for(std::size_t ie = 0; ie < num_cells; ++ie){
        int nm1 = cells[ie*3 + 0];
        int nm2 = cells[ie*3 + 1];
        int nm3 = cells[ie*3 + 2];

        for(std::size_t j = 0; j < 2; ++j){
            tmp[kount] = points[2*nm1+j];
            kount += 1;
        }
        for(std::size_t j = 0; j < 2; ++j){
            tmp[kount] = points[2*nm2+j];
            kount += 1;
        }
        for(std::size_t j = 0; j < 2; ++j){
            tmp[kount] = points[2*nm3+j];
            kount += 1;
        }
    }
    std::vector<double> dist = c_drectangle(tmp, x1, x2, y1, y2);
    return dist;
}

py::array remove_external_entities2(
    py::array_t<double, py::array::c_style | py::array::forcecast> points,
    py::array_t<int, py::array::c_style | py::array::forcecast> cells,
    double x1, double x2, double y1, double y2) {

  // check input dimensions
  int num_points = points.size() / 2;
  int num_cells  = cells.size() / 3;

  // allocate std::vector (to pass to the C++ function)
  std::vector<double> cppPts(num_points * 2);
  std::vector<int> cppCells(num_cells * 3);

  // copy py::array -> std::vector
  std::memcpy(cppPts.data(), points.data(), 2 * num_points * sizeof(double));
  std::memcpy(cppCells.data(), cells.data(), 3 * num_cells * sizeof(int));

  std::vector<double> dist = c_remove_external_entities2(cppPts, cppCells, x1, x2, y1, y2);

  ssize_t sodble = sizeof(double);
  std::vector<ssize_t> shape = {num_cells*3};
  std::vector<ssize_t> strides = {sodble};

  // return 1-D NumPy array
  return py::array(
      py::buffer_info(dist.data(),    /* data as contiguous array  */
                      sizeof(double), /* size of one scalar        */
                      py::format_descriptor<double>::format(), /* data type */
                      1,      /* number of dimensions      */
                      shape,  /* shape of the matrix       */
                      strides /* strides for each axis     */
                      ));
}


std::vector<double> c_remove_external_entities3(std::vector<double> &points, std::vector<int> &cells, double x1, double x2, double y1, double y2, double z1, double z2){

    int num_points = points.size()/3;
    int num_cells = cells.size()/4;
    std::vector<double> tmp;
    tmp.resize(num_cells*4*3);

    int kount = 0;
    for(std::size_t ie = 0; ie < num_cells; ++ie){
        int nm1 = cells[ie*4 + 0];
        int nm2 = cells[ie*4 + 1];
        int nm3 = cells[ie*4 + 2];
        int nm4 = cells[ie*4 + 3];

        for(std::size_t j = 0; j < 3; ++j){
            tmp[kount] = points[3*nm1+j];
            kount += 1;
        }
        for(std::size_t j = 0; j < 3; ++j){
            tmp[kount] = points[3*nm2+j];
            kount += 1;
        }
        for(std::size_t j = 0; j < 3; ++j){
            tmp[kount] = points[3*nm3+j];
            kount += 1;
        }
        for(std::size_t j = 0; j < 3; ++j){
            tmp[kount] = points[3*nm4+j];
            kount += 1;
        }
    }
    std::vector<double> dist = c_dblock(tmp, x1, x2, y1, y2, z1, z2);
    return dist;
}

py::array remove_external_entities3(
    py::array_t<double, py::array::c_style | py::array::forcecast> points,
    py::array_t<int, py::array::c_style | py::array::forcecast> cells,
    double x1, double x2, double y1, double y2, double z1, double z2) {

  // check input dimensions
  int num_points = points.size() / 3;
  int num_cells  = cells.size() / 4;

  // allocate std::vector (to pass to the C++ function)
  std::vector<double> cppPts(num_points * 3);
  std::vector<int> cppCells(num_cells * 4);

  // copy py::array -> std::vector
  std::memcpy(cppPts.data(), points.data(), 3 * num_points * sizeof(double));
  std::memcpy(cppCells.data(), cells.data(), 4 * num_cells * sizeof(int));

  std::vector<double> dist = c_remove_external_entities3(cppPts, cppCells, x1, x2, y1, y2, z1, z2);

  ssize_t sodble = sizeof(double);
  std::vector<ssize_t> shape = {num_cells*4};
  std::vector<ssize_t> strides = {sodble};

  // return 1-D NumPy array
  return py::array(
      py::buffer_info(dist.data(),    /* data as contiguous array  */
                      sizeof(double), /* size of one scalar        */
                      py::format_descriptor<double>::format(), /* data type */
                      1,      /* number of dimensions      */
                      shape,  /* shape of the matrix       */
                      strides /* strides for each axis     */
                      ));
}


std::vector<double> c_calc_dihedral_angles(std::vector<double> &points,
                                           std::vector<int> &cells) {
  // compute the 6 dihedral angles of all tetrahedrons in the mesh
  static const std::size_t edges[6][2] = {{2, 3}, {1, 3}, {1, 2},
                                          {0, 3}, {0, 2}, {0, 1}};

  int num_cells = cells.size() / 4;

  std::vector<double> dh_angles;
  dh_angles.resize(num_cells * 6);

  std::array<double, 3> p0;
  std::array<double, 3> v1;
  std::array<double, 3> v2;
  std::array<double, 3> v3;

  for (unsigned int c = 0; c < num_cells; ++c) {
    for (unsigned int i = 0; i < 6; ++i) {
      const std::size_t i0 = cells[4 * c + edges[i][0]];
      const std::size_t i1 = cells[4 * c + edges[i][1]];
      const std::size_t i2 = cells[4 * c + edges[5 - i][0]];
      const std::size_t i3 = cells[4 * c + edges[5 - i][1]];

      // populate point coordinates
      for (unsigned int j = 0; j < 3; ++j) {
        p0[j] = points[i0 * 3 + j];
        v1[j] = points[i1 * 3 + j];
        v2[j] = points[i2 * 3 + j];
        v3[j] = points[i3 * 3 + j];
      }

      // subtract p0 vector
      for (unsigned int j = 0; j < 3; ++j) {
        v1[j] -= p0[j];
        v2[j] -= p0[j];
        v3[j] -= p0[j];
      }

      // normalize by l2-norm
      const double v1_div = l2_norm(v1);
      const double v2_div = l2_norm(v2);
      const double v3_div = l2_norm(v3);

      for (unsigned int j = 0; j < 3; ++j) {
        v1[j] /= v1_div;
        v2[j] /= v2_div;
        v3[j] /= v3_div;
      }

      // compute dot product
      double v2Dotv3 = dot_product(v2, v3);
      double v1Dotv2 = dot_product(v1, v2);
      double v1Dotv3 = dot_product(v1, v3);

      // compute cross product
      std::array<double, 3> v1Crossv2 = cross_product(v1, v2);
      std::array<double, 3> v1Crossv3 = cross_product(v1, v3);

      const double norm_v1Crossv2 = l2_norm(v1Crossv2);
      const double norm_v1Crossv3 = l2_norm(v1Crossv3);

      double cphi =
          (v2Dotv3 - v1Dotv2 * v1Dotv3) / ((norm_v1Crossv2 * norm_v1Crossv3));
      dh_angles[c * 6 + i] = std::acos(cphi); // in radians
    }
  }
  return dh_angles;
}

py::array calc_dihedral_angles(
    py::array_t<double, py::array::c_style | py::array::forcecast> pts,
    py::array_t<int, py::array::c_style | py::array::forcecast> cells) {

  // check input dimensions
  int num_points = pts.size() / 3;
  int num_cells = cells.size() / 4;
  // allocate std::vector (to pass to the C++ function)
  std::vector<double> cppPts(num_points * 3);
  std::vector<int> cppCells(num_cells * 4);

  // copy py::array -> std::vector
  std::memcpy(cppPts.data(), pts.data(), 3 * num_points * sizeof(double));
  std::memcpy(cppCells.data(), cells.data(), 4 * num_cells * sizeof(int));

  std::vector<double> dh_angles = c_calc_dihedral_angles(cppPts, cppCells);

  ssize_t num_angles = num_cells * 6;
  ssize_t ndim = 2;
  ssize_t sodble = sizeof(double);
  std::vector<ssize_t> shape = {num_angles, 1};
  std::vector<ssize_t> strides = {sodble, sodble};

  // return 2-D NumPy array
  return py::array(
      py::buffer_info(dh_angles.data(), /* data as contiguous array  */
                      sizeof(double),   /* size of one scalar        */
                      py::format_descriptor<double>::format(), /* data type */
                      2,      /* number of dimensions      */
                      shape,  /* shape of the matrix       */
                      strides /* strides for each axis     */
                      ));
}

// fixed size calculation for 3x3 determinant
double c_calc_3x3determinant(std::vector<double> &m) {
  // | (0,0) 0  (0,1) 1 (0,2) 2|
  // | (1,0) 3  (1,1) 4 (1,2) 5|
  // | (2,0) 6  (2,1) 7 (2,2) 8|
  return m[0] * (m[4] * m[8] - m[7] * m[5]) -
         m[1] * (m[3] * m[8] - m[6] * m[5]) +
         m[2] * (m[3] * m[7] - m[6] * m[4]);
}

// ----------------
// Python interface for c_calc_3x3determinant
// ----------------
double calc_3x3determinant(
    py::array_t<double, py::array::c_style | py::array::forcecast> matrix) {
  // allocate std::vector (to pass to the C++ function)
  std::vector<double> cppMatrix(9);

  // copy py::array -> std::vector
  std::memcpy(cppMatrix.data(), matrix.data(), 9 * sizeof(double));

  // call cpp code
  double result = c_calc_3x3determinant(cppMatrix);

  return result;
}

// fixed size calculation for 4x4 determinant
double c_calc_4x4determinant(std::vector<double> &m) {
  return m[12] * m[9] * m[6] * m[3] - m[8] * m[13] * m[6] * m[3] -
         m[12] * m[5] * m[10] * m[3] + m[4] * m[13] * m[10] * m[3] +
         m[8] * m[5] * m[14] * m[3] - m[4] * m[9] * m[14] * m[3] -
         m[12] * m[9] * m[2] * m[7] + m[8] * m[13] * m[2] * m[7] +
         m[12] * m[1] * m[10] * m[7] - m[0] * m[13] * m[10] * m[7] -
         m[8] * m[1] * m[14] * m[7] + m[0] * m[9] * m[14] * m[7] +
         m[12] * m[5] * m[2] * m[11] - m[4] * m[13] * m[2] * m[11] -
         m[12] * m[1] * m[6] * m[11] + m[0] * m[13] * m[6] * m[11] +
         m[4] * m[1] * m[14] * m[11] - m[0] * m[5] * m[14] * m[11] -
         m[8] * m[5] * m[2] * m[15] + m[4] * m[9] * m[2] * m[15] +
         m[8] * m[1] * m[6] * m[15] - m[0] * m[9] * m[6] * m[15] -
         m[4] * m[1] * m[10] * m[15] + m[0] * m[5] * m[10] * m[15];
}

// ----------------
// Python interface for c_calc_4x4determinant
// ----------------
double calc_4x4determinant(
    py::array_t<double, py::array::c_style | py::array::forcecast> matrix) {
  // allocate std::vector (to pass to the C++ function)
  std::vector<double> cppMatrix(16);

  // copy py::array -> std::vector
  std::memcpy(cppMatrix.data(), matrix.data(), 16 * sizeof(double));

  // call cpp code
  double result = c_calc_4x4determinant(cppMatrix);

  return result;
}

// Calculate the gradient of the tetrahedral's volume wrt to point0
std::vector<double> c_calc_volume_grad(std::vector<double> &p1,
                                       std::vector<double> &p2,
                                       std::vector<double> &p3) {
  int num_points = p1.size() / 3;
  double x1, x2, x3;
  double y1, y2, y3;
  double z1, z2, z3;
  std::vector<double> gradient;
  gradient.resize(3 * num_points);
  for (int i = 0; i < num_points; ++i) {
    // unpack coordinates
    x1 = p1[3 * i];
    y1 = p1[3 * i + 1];
    z1 = p1[3 * i + 2];
    x2 = p2[3 * i];
    y2 = p2[3 * i + 1];
    z2 = p2[3 * i + 2];
    x3 = p3[3 * i];
    y3 = p3[3 * i + 1];
    z3 = p3[3 * i + 2];
    //
    gradient[3 * i] =
        (1.0 / 6.0) * (y2 * z3 + y1 * (z2 - z3) - y3 * z2 - z1 * (y2 - y3));
    gradient[3 * i + 1] =
        (1.0 / 6.0) * (-x2 * z3 - x1 * (z2 - z3) + x3 * z2 + z1 * (x2 - x3));
    gradient[3 * i + 2] =
        (1.0 / 6.0) * (x2 * y3 + x1 * (y2 - y3) - x3 * y2 - y1 * (x2 - x3));
  }
  return gradient;
}

// Python wrapper accepts the coordinates of the slivers' vertices
py::array calc_volume_grad(
    py::array_t<double, py::array::c_style | py::array::forcecast> p1,
    py::array_t<double, py::array::c_style | py::array::forcecast> p2,
    py::array_t<double, py::array::c_style | py::array::forcecast> p3) {

  int num_points = p1.size() / 3;

  std::vector<double> cppP1(3 * num_points);
  std::vector<double> cppP2(3 * num_points);
  std::vector<double> cppP3(3 * num_points);

  std::memcpy(cppP1.data(), p1.data(), 3 * num_points * sizeof(double));
  std::memcpy(cppP2.data(), p2.data(), 3 * num_points * sizeof(double));
  std::memcpy(cppP3.data(), p3.data(), 3 * num_points * sizeof(double));

  std::vector<double> volume_grad = c_calc_volume_grad(cppP1, cppP2, cppP3);

  ssize_t sodble = sizeof(double);
  std::vector<ssize_t> shape = {num_points, 3};
  std::vector<ssize_t> strides = {sodble * 3, sodble};

  // return 2-D NumPy array
  return py::array(
      py::buffer_info(volume_grad.data(), /* data as contiguous array  */
                      sizeof(double),     /* size of one scalar        */
                      py::format_descriptor<double>::format(), /* data type */
                      2,      /* number of dimensions      */
                      shape,  /* shape of the matrix       */
                      strides /* strides for each axis     */
                      ));
}

// used to guide the pertubation to the point
std::vector<double> c_calc_circumsphere_grad(std::vector<double> &p0,
                                             std::vector<double> &p1,
                                             std::vector<double> &p2,
                                             std::vector<double> &p3) {
  const int num_points = p0.size() / 3;

  // coordinates of cell
  double x1;
  double y1;
  double z1;

  double x2;
  double y2;
  double z2;

  double x3;
  double y3;
  double z3;

  std::vector<double> gradient_cradius;
  gradient_cradius.resize(num_points * 3);

  for (int i = 0; i < num_points; ++i) {
    // translate the tet so that fourth point is the origin
    x1 = p0[3 * i] - p3[3 * i];
    y1 = p0[3 * i + 1] - p3[3 * i + 1];
    z1 = p0[3 * i + 2] - p3[3 * i + 2];

    x2 = p1[3 * i] - p3[3 * i];
    y2 = p1[3 * i + 1] - p3[3 * i + 1];
    z2 = p1[3 * i + 2] - p3[3 * i + 2];

    x3 = p2[3 * i] - p3[3 * i];
    y3 = p2[3 * i + 1] - p3[3 * i + 1];
    z3 = p2[3 * i + 2] - p3[3 * i + 2];

    // pre-compute everything
    double sq_p1 = x1 * x1 + y1 * y1 + z1 * z1;
    double sq_p2 = x2 * x2 + y2 * y2 + z2 * z2;
    double sq_p3 = x3 * x3 + y3 * y3 + z3 * z3;

    // every derivative is computed w.r.t p1 (x1, y1, z1)
    double da_dx = y2 * z3 - y3 * z2;
    double da_dy = z2 * x3 - x2 * z3;
    double da_dz = x2 * y3 - x3 * y2;

    double dDx_dx = -2.0 * x1 * da_dx;
    double dDx_dy = -2.0 * y1 * da_dx + sq_p2 * z3 - sq_p3 * z2;
    double dDx_dz = -2.0 * z1 * da_dx - sq_p2 * y3 + sq_p3 * y2;

    double dDy_dx = -2.0 * x1 * da_dy - sq_p2 * z3 + sq_p3 * z2;
    double dDy_dy = -2.0 * y1 * da_dy;
    double dDy_dz = -2.0 * z1 * da_dy + sq_p2 * x3 - sq_p3 * x2;

    double dDz_dx = -2.0 * x1 * da_dz + sq_p2 * y3 - sq_p3 * y2;
    double dDz_dy = -2.0 * y1 * da_dz - sq_p2 * x3 + sq_p3 * x2;
    double dDz_dz = -2.0 * z1 * da_dz;

    double a = x1 * da_dx + y1 * da_dy + z1 * da_dz;
    // std::cout<<a<<std::endl;
    // if ( CGAL_NTS is_zero(a) )
    //  return CGAL::NULL_VECTOR;

    double Dx = -sq_p1 * da_dx + y1 * (sq_p2 * z3 - sq_p3 * z2) -
                z1 * (sq_p2 * y3 - sq_p3 * y2);
    double Dy = -sq_p1 * da_dy - x1 * (sq_p2 * z3 - sq_p3 * z2) +
                z1 * (sq_p2 * x3 - sq_p3 * x2);
    double Dz = -sq_p1 * da_dz + x1 * (sq_p2 * y3 - sq_p3 * y2) -
                y1 * (sq_p2 * x3 - sq_p3 * x2);

    // compute gradient vector
    double sum_sqD = Dx * Dx + Dy * Dy + Dz * Dz;
    double gx = (Dx * dDx_dx + Dy * dDy_dx + Dz * dDz_dx) / (2.0 * a * a) -
                (da_dx * sum_sqD) / (2.0 * a * a * a);
    double gy = (Dx * dDx_dy + Dy * dDy_dy + Dz * dDz_dy) / (2.0 * a * a) -
                (da_dy * sum_sqD) / (2.0 * a * a * a);
    double gz = (Dx * dDx_dz + Dy * dDy_dz + Dz * dDz_dz) / (2.0 * a * a) -
                (da_dz * sum_sqD) / (2.0 * a * a * a);

    gradient_cradius[3 * i] = gx;
    gradient_cradius[3 * i + 1] = gy;
    gradient_cradius[3 * i + 2] = gz;
  }

  // the square of the point coordinates
  return gradient_cradius;
}

// Python wrapper accepts the coordinates of the slivers' vertices
py::array calc_circumsphere_grad(
    py::array_t<double, py::array::c_style | py::array::forcecast> p0,
    py::array_t<double, py::array::c_style | py::array::forcecast> p1,
    py::array_t<double, py::array::c_style | py::array::forcecast> p2,
    py::array_t<double, py::array::c_style | py::array::forcecast> p3) {

  int num_points = p0.size() / 3;

  std::vector<double> cppP0(3 * num_points);
  std::vector<double> cppP1(3 * num_points);
  std::vector<double> cppP2(3 * num_points);
  std::vector<double> cppP3(3 * num_points);

  std::memcpy(cppP0.data(), p0.data(), 3 * num_points * sizeof(double));
  std::memcpy(cppP1.data(), p1.data(), 3 * num_points * sizeof(double));
  std::memcpy(cppP2.data(), p2.data(), 3 * num_points * sizeof(double));
  std::memcpy(cppP3.data(), p3.data(), 3 * num_points * sizeof(double));

  std::vector<double> circumsphere_grad =
      c_calc_circumsphere_grad(cppP0, cppP1, cppP2, cppP3);

  ssize_t sodble = sizeof(double);
  std::vector<ssize_t> shape = {num_points, 3};
  std::vector<ssize_t> strides = {sodble * 3, sodble};

  // return 2-D NumPy array
  return py::array(
      py::buffer_info(circumsphere_grad.data(), /* data as contiguous array  */
                      sizeof(double),           /* size of one scalar        */
                      py::format_descriptor<double>::format(), /* data type */
                      2,      /* number of dimensions      */
                      shape,  /* shape of the matrix       */
                      strides /* strides for each axis     */
                      ));
}

PYBIND11_MODULE(fast_geometry, m) {
  m.def("remove_external_entities2", &remove_external_entities2);
  m.def("remove_external_entities3", &remove_external_entities3);
  m.def("drectangle_fast", &drectangle_fast);
  m.def("dblock_fast", &dblock_fast);
  m.def("unique_edges", &unique_edges);
  m.def("calc_volume_grad", &calc_volume_grad);
  m.def("calc_circumsphere_grad", &calc_circumsphere_grad);
  m.def("calc_dihedral_angles", &calc_dihedral_angles);
  m.def("calc_4x4determinant", &calc_4x4determinant);
  m.def("calc_3x3determinant", &calc_3x3determinant);
}
